/*
This is a Optical-Character-Recognition program
Copyright (C) 2000  Joerg Schulenburg

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 Joerg.Schulenburg@physik.uni-magdeburg.de
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pgm2asc.h"
#include "gocr.h"
#include "pnm.h"

/* FIXME jb global */
int debug=0; /* its only for development phase */

/*
     detect barcode and add a string to the box (obj-pointer, ToDo)
     ToDo: barcode-output stderr->stdout
*/

/* "BSBSBS", B=Bar, S=Space, better using 2*6=12bit-integer? */
#define Num128 107
const char *code128[Num128]={ /* can be generated by an algorithm? */
/* 00 */"212222","222122","222221","121223","121322","131222","122213","122312",
/* 08 */"132212","221213","221312","231212","112232","122132","122231","113222",
/* 16 */"123122","123221","223211","221132","221231","213212","223112","312131",
/* 24 */"311222","321122","321221","312212","322112","322211","212123","212321",
/* 32 */"232121","111323","131123","131321","112313","132113","132311","211313",
/* 40 */"231113","231311","112133","112331","132131","113123","113321","133121",
/* 48 */"313121","211331","231131","213113","213311","213131","311123","311321",
/* 56 */"331121","312113","312311","332111","314111","221411","431111","111224",
/* 64 */"111422","121124","121421","141122","141221","112214","112412","122114",
/* 72 */"122411","142112","142211","241211","221114","413111","241112","134111",
/* 80 */"111242","121142","121241","114212","124112","124211","411212","421112",
/* 88 */"421211","212141","214121","412121","111143","111341","131141","114113",
/* 96 */"114311","411113","411311","113141","114131","311141","411131","211412",
/*104 */"211214","211232","2331112"};

/*
      code128: see code128.tex by Petr Olsak (108 codes)
       quiet_zone: size=10 (before and after code128)
       num_bars=3*(star+chars[N]+crc+stop)+1
       char=bar,space,bar,space,bar,space (size=11), stop=4bars (size=11+2)
       Width: Bar,Space=1,[2,3,4] char=11 code=11*(N+3)+2 sumB=even,sumS=odd
       startA=211412 startB=211214 startC=211232(2dec_digits)
       crc=(start+1*char1+2*char2+3*char3+...+N*charN) mod 103
       stop=2331112 (4bars)  => start/end==211
       return num of chars or string
       
       size  B+W    even-variants      odd-variants      num_codes 
        11 = 8+3 = (1+3+4,2+2+4,2+3+3)+(1+1+1)       => (6+3+3)*(1)   =  12
           = 6+5 = (1+1+4,1+2+3,2+2+2)+(1+1+3,1+2+2) => (3+6+1)*(3+3) =  60
           = 4+7 = (1+1+2)+(1+2+4,1+3+3,2+2+3)       => (3)*(6+3+3)   =  36
                                                                  sum = 108
 */

int decode_code128(int *wb, int num_bars){ /* ToDo: char *dest, int len */
  int i, w, dd, i1, i2, i3, i4, i5, db, ds, dx, crc, mode=0;
  double dw;  char cc, buf[8];
  for(w=i=0;i<2*num_bars-1;i++) w+=wb[i];
  
  /* test code128 characteristics, ToDo: look for correct start/stop 211 seq. */
  if ((num_bars-1)%3!=0 || num_bars<10 || w<11*(num_bars-1)/3+2)
    return 0;
  if (2*wb[0]<3*wb[1] || 2*wb[0]<3*wb[2]) return 0; /* 211 */
  dw=3.0*w/((num_bars-1)*11+6);
  fprintf(stderr," code128 base=%.2f chars=%2d ",dw,(num_bars-1)/3-3);
  /* ToDo: find range of 1=14..18 2=26..33 3=39..50 */
  crc=0;
  for(i1=0;i1<(num_bars-1)/3;i1++) {
    for(i3=i2=0;i2<6;i2++) { i3+=wb[6*i1+i2]; } /* better results */
    dw=i3/11.0;
    for(db=ds=i2=0;i2<6;i2++) {
      dx=wb[6*i1+i2];
      dd=(int)(dx/dw+0.25+0.5*(i2&1)); /* ToDo: if not exact try dx+1, dx-1 */
      if (debug) fprintf(stderr,"\n %c dx= %3d %.4f %2d ",
                        ((i2&1)?'S':'B'),dx,dx/dw,dd);
      if (i2&1) ds+=dd;
      else      db+=dd;
      buf[i2]='0'+dd;
    } buf[6]='\0'; /* string end */
    if (debug) fprintf(stderr, "len= %3d %d+%d ",i3,db,ds);
    i3=Num128;
    if (db+ds==11 && db%2==0 && ds%2==1) /* looks well */
      for(i3=0;i3<Num128;i3++) if (strstr(code128[i3],buf)) break;
    /* primitive error correction */
    if (db+ds!=11 || db%2!=0 || ds%2!=1) { /* ToDo: error corrections */
      /* simplest way: find best accordance */
      for(i5=0;i5<Num128;i5++){
        for(i4=i2=0;i2<6;i2++)
          if (code128[i5][i2]==buf[i2]) i4++;
        if(i4>=5) break;
      }
    }

    if(i3<Num128){  /* valid symbol */
      if(i1==0){
        mode=((i3>102 && i3<106)?i3-103:-1);
        crc=i3;
        if(mode>=0) fprintf(stderr," start%c ",'A'+mode);
        else        fprintf(stderr," %s=%02d ",buf,i3);
      }
      if(i1>0 && i1<(num_bars-1)/3-2){
        crc+=i3*i1; cc='\0';
        if (mode==0 && i3<64) cc=' '+i3; 
        if (mode==1 && i3<95) cc=' '+i3; 
        if (cc) fprintf(stderr,"%c",cc);
        else    fprintf(stderr,",%02d",i3);
      }
      if(i1==(num_bars-1)/3-2){
        fprintf(stderr," crc:%02d=%02d",crc%103,i3);
      }
      if(i1==(num_bars-1)/3-1){
        if(i3==106) fprintf(stderr," stop=%02d",i3);
        else        fprintf(stderr," %s=%02d",buf,i3);
      }
    }
    else fprintf(stderr," %s=%02d? ",buf,i5);
  }
  return 0;
}

#define NumUPC 20
const char *codeUPC[NumUPC]={ /* 0..9 left SBSB, right BSBS even/odd */
  "3211","2221","2122","1411","1132","1231","1114","1312","1213","3112",
  "1123","1222","2212","1141","2311","1321","4111","2131","3121","2113"}; /* inverse! */

/*
      EAN 13 (UPC,(1+6+1+6+1)*2bars,size=3+6*7+5+6*7+3=95) 
      EAN  8 (UPC,(1+4+1+4+1)*2bars,size=3+4*7+5+4*7+3=67)
      UPC: (10 codes)
       num_bars=2*(2*6+3) middle=SBSBS=11111 right/left=BSB=111="101"
       char: left=SBSB right=BSBS (size=7) only_dec_digits
      size ev+odd  even  + odd
       7 = 2 + 5 = (1+1) + (1+4,2+3)   => (1)*(2+2) =  4 codes
         = 4 + 3 = (1+3,2+2) + (1+3)   => (2+1)*(2) =  6 codes
                                                sum = 10 codes 
 */

int decode_UPC(int *wb, int num_bars){ /* ToDo: char *dest, int len */
  int i, w, i1, i2, i3, i4, i5, dd, db, ds, dx, crc;
  double dw;  char cc, buf[8];
  for(w=i=0;i<2*num_bars-1;i++) w+=wb[i];

  dw=2.0*w/((num_bars-6)*7+2*11); /* or min(wb[]) */
  crc=0;
  if ((num_bars)%2!=0 || num_bars<10 || w<7*(num_bars-6)/2+11
   || ((num_bars-6)/2)%2!=0) return 0;  /* should be balanced */
  /* check for correct start symbol BSB or 111 */
  i1=wb[0]; if (i1>wb[1]) i1=wb[1]; if (i1>wb[2]) i1=wb[2]; /* min */
  i2=wb[0]; if (i2<wb[1]) i2=wb[1]; if (i2<wb[2]) i2=wb[2]; /* max */
  if (2*wb[0]>3*i1 || 2*wb[1]>3*i1 || 2*wb[2]>3*i1) return 0;
  if (3*wb[0]<2*i2 || 3*wb[1]<2*i2 || 3*wb[2]<2*i2) return 0;
  /* ToDo: check middle symbol SBSBS or 11111 */
  /* ToDo: check for correct end symbol BSB or 111 */
  i=2*num_bars-2;
  if (2*wb[i-0]>3*i1 || 2*wb[i-1]>3*i1 || 2*wb[i-2]>3*i1) return 0;
  if (3*wb[i-0]<2*i2 || 3*wb[i-1]<2*i2 || 3*wb[i-2]<2*i2) return 0;
  fprintf(stderr," UPC     base=%.2f chars=%2d ",dw,(num_bars-6)/2);

  for(i1=3;i1<2*num_bars-4;i1+=4) { /* 1st half: SBSB, 2nd: BSBS */
    i4=i1/(num_bars*2); /* 0=left or 1=right side */
    for(db=ds=i2=0;i2<4;i2++) {
      i5=(i1+i2)&1;       /* 0=Bar or 1=Space */
      dx=wb[i1+i2];
      dd=(int)(dx/dw+0.5); /* ToDo: if not exact try dx+1, dx-1 */
      if (debug) fprintf(stderr,"\n %c %3d dx=%3d %.4f %2d ",
                        ((i5)?'B':'S'),i1,dx,dx/dw,dd);
      if (i5) db+=dd;
      else    ds+=dd;
      buf[i2]='0'+dd;
    } buf[4]='\0'; /* string end */
    if (debug) fprintf(stderr," %d+%d ",db,ds);
    if (strstr("1111",buf)) { i1++; continue; } /* skip middle sync */
    i3=NumUPC;
    if (db+ds==7 && ((!i4) || (db%2==1 && ds%2==0))) /* looks well */
      for(i3=0;i3<NumUPC;i3++) if (strstr(codeUPC[i3],buf)) break;

    if(i3<NumUPC){  /* valid symbol */
        fprintf(stderr,"%1d",i3%10);
    }
    else
    { /* ToDo: error corrections */
      /* simplest way: find best accordance */
      for(i3=0;i3<NumUPC;i3++){
        for(i4=i2=0;i2<4;i2++)
          if (codeUPC[i3][i2]==buf[i2]) i4++;
        if(i4>=3) break;
      }
       fprintf(stderr," %s=%d? ",buf,i3%10);
    }
  }

  return 0;
}

/* code 3 of 9, BSBSBSBSB<S> size=6+3*x x=2..3 */
#define Num39 (40+4) /* (2of5)(1of4)+(0of5)(3of4), (2of5)(.-..)=0..9 */
const char *code39= /* rearranged to BBBBBSSSS<S> */
"1-...-.-..""2.-..-.-..""3--....-..""4..-.-.-..""5-.-...-.."
"6.--...-..""7...--.-..""8-..-..-..""9.-.-..-..""0..--..-.."
"A-...-..-.""B.-..-..-.""C--.....-.""D..-.-..-.""E-.-....-."
"F.--....-.""G...--..-.""H-..-...-.""I.-.-...-.""J..--...-."
"K-...-...-""L.-..-...-""M--......-""N..-.-...-""O-.-.....-"
"P.--.....-""Q...--...-""R-..-....-""S.-.-....-""T..--....-"
"U-...--...""V.-..--...""W--...-...""X..-.--...""Y-.-..-..."
"Z.--..-...""-...---..."".-..-.-..."" .-.-.-...""*..--.-..."
/* (0of5)(3of4)=(.....)(3of4) store only 3of4? */
"$.....---.""/.....--.-""+.....-.--""%......---";

/* code interleaved 2 of 5
   B B B B B  <= digit 1,3,5,...
    S S S S S <= digit 2,4,6,...
    1 2 4 7 0 <= weight   digit=sum(bit*weight)%11
 */
const char *code25= /* is the code sorted randomly? */
"1-...-2.-..-3--...4..-.-5-.-..6.--..7...--8-..-.9.-.-.0..--.";

/* not tested */
int decode_39(int *wb, int num_bars){ /* ToDo: char *dest, int len */
  int i, w, i1, i2, i3, i4, i5, dd, db, ds, dx, crc;
  double dw;  char cc, buf[12];
  for(w=i=0;i<2*num_bars-1;i++) w+=wb[i];

  for (dw=w,i=0;i<num_bars-1;i++) if (dw>wb[i]) dw=wb[i];
  dw=dw*3/2;
  /* change dw in a way that we find 3 of 9 thick */
  for(;;){
    /* count thick bars/spaces */
    for (i1=i=0;i<num_bars-1;i++) if (dw<=wb[i]) i1++;
    if (i1>3*num_bars/5) dw++;
    else break;
  }
  for(;;){
    /* count thick bars/spaces */
    for (i1=i=0;i<num_bars-1;i++) if (dw<=wb[i]) i1++;
    if (i1<3*num_bars/5) dw--;
    else break;
  }
  crc=0;
  if ((num_bars)%5!=0 || num_bars<15) return 0; /* chars*5bars */
  /* check for correct start and end symbol * or NwNnWnWnN Narrow+Wide */
  i=2*num_bars-2;
  // if (wb[  0]>dw || wb[  1]<=dw || wb[  2]> dw) return 0;
  // if (wb[i-0]>dw || wb[i-1]> dw || wb[i-2]<=dw) return 0;
  fprintf(stderr," code39  base=%.2f chars=%2d ",dw,(num_bars)/5);

  for(i1=0;i1<2*num_bars-3;i1+=10) { /* 1st half: SBSB, 2nd: BSBS */
    /* ToDo: looking for three widest bars/gaps */
    for(db=ds=i2=0;i2<9;i2++) {
      i5=(i2&1);       /* 0=Bar or 1=Space */
      dx=wb[i1+i2];
      dd=((2*dx<3*dw)?0:1);
      if (debug) fprintf(stderr,"\n %c=dx,dx/dw,dd=%3d %.4f %2d ",
                        ((i5)?'S':'B'),dx,dx/dw,dd);
      if (i5) db+=dd;
      else    ds+=dd;
      buf[(i2%2)*5+i2/2]=(dd)?'-':'.'; /* BBBBBSSSS */
    } buf[9]='\0'; /* string end */
    i3=Num39;
    if (db+ds==3) /* looks well */
      for(i3=0;i3<Num39;i3++)
        if (strstr(&code39[10*i3+1],buf)==&code39[10*i3+1]) break;

    if (debug) fprintf(stderr," %s ",buf);
    if(i3<Num39){  /* valid symbol */
        fprintf(stderr,"%c",code39[10*i3]);
    }
    else
    { /* ToDo: error corrections */
       fprintf(stderr," %s=? ",buf);
    }
  }

  return 0;
}

/*
    decode barcode
    - check main characteristics (num bars, min+max width, etc.)
    - detect code type
    - the current algorithm measures the width of bars and spaces
    - probably its better to look at (n+0.5)*base
 */

int decode_barcode(struct box *bb){ /* ToDo: char *dest, int len */
  int i, num_bars, yy, w, ww, dx, xx, cs=JOB->cfg.cs, *wb;
  yy=(bb->y0+bb->y1)/2;
  w=ww=bb->x1-bb->x0+1;
  num_bars = num_cross(bb->x0,bb->x1,yy,yy,bb->p,JOB->cfg.cs);
  fprintf(stderr," bars=%3d w=%4d",num_bars,ww);
  
  /* store width of bars and spaces to buffer wb */
  wb=malloc(2*num_bars*sizeof(int)); if(!wb) return -1;
  xx=bb->x0;
  xx-=loop(bb->p,xx,yy, 8,cs,1,LE);
  xx+=loop(bb->p,xx,yy,ww,cs,0,RI); /* start with a bar! */
  for (i=0;i<2*num_bars;i++) {
    dx=loop(bb->p,xx,yy,w,cs,1^(i&1),RI);
    xx+=dx;
    w-=dx;
    wb[i]=dx;
  } wb[2*num_bars-1]=0;
  
  /* test code128 characteristics, ToDo: look for correct start/stop 211 seq. */
  if ((num_bars-1)%3==0 && num_bars>=10 && ww>=11*(num_bars-1)/3+2){
    decode_code128(wb,num_bars);
  }
  /* test UPC/EAN characteristics (not ready) */
  if ((num_bars)%2==0 && num_bars>=8 && ww>=7*(num_bars-6)/2+11
  && ((num_bars-6)/2)%2==0){  /* should be balanced */
    decode_UPC(wb,num_bars);
  }
  /* test code39 characteristics (not ready) */
  if ((num_bars)%5==0 && num_bars>14){
    decode_39(wb,num_bars);
  }
  free(wb);
  return 0;
}

int detect_barcode(job_t *job)
{
  // FIXME jb: remove p
  pix *p = &job->src.p;
  int i, j, j2=0, bx0, by0, bx1, by1, bdx, bdy, bbx,
     x0, y0, dx, dy, cs, y, yl0, yl1, yr0, yr1;
  struct box *box2, *box3;

  x0=0;  y0=0;  dx=p->x;  dy=p->y; cs=JOB->cfg.cs;
  for_each_data(&(JOB->res.boxlist)) {
    box2 = (struct box *)list_get_current(&(JOB->res.boxlist));
    bdx=box2->x1-box2->x0+1
       -loop(p,box2->x1,(box2->y0+box2->y1)/2,box2->x1-box2->x0,cs,0,LE)
       -loop(p,box2->x0,(box2->y0+box2->y1)/2,box2->x1-box2->x0,cs,0,RI);
    bdy=box2->y1-box2->y0+1;
    if (box2->c == PICTURE || box2->c == UNKNOWN)
    if (box2->y0 >= y0 && box2->y1 <= y0 + dy  /* within frame? */
     && box2->x0 >= x0 && box2->x1 <= x0 + dx
     && box2->y1 - box2->y0 > 19               /* min. length */
     && box2->y1 - box2->y0 > 16 * bdx
     && (get_bw(box2->x0+bdx/8 ,box2->x1-bdx/8,
                box2->y0+bdy/16,box2->y1-bdy/16,box2->p,cs+20,2)==0
                 ||     bdy > 8*bdx)
    ) { /* a bar? */
       j=1; /* number of bars */
       bx0=box2->x0; bx1=box2->x1;
       by0=box2->y0; by1=box2->y1;
       bbx=bx1-bx0+2; /* width of bar */
       /* this is for scans which are not exactly horizontal */
       yl0=yr0=by0; /* left and right upper bound */
       yl1=yr1=by1; /* left and right lower bound */
       j2=1; /* flag if bar added */
       while(j2) { /* repeat searching until all bars added */
         j2=0;
         for_each_data(&(JOB->res.boxlist)) {
           box3 = (struct box *)list_get_current(&(JOB->res.boxlist));
           bdy=box3->y1-box3->y0+1;
           if (box2!=box3)
           if (box3->c == PICTURE || box3->c == UNKNOWN)
           if (box3->y0 >= y0 && box3->y1 <= y0 + dy  /* within frame? */
            && box3->x0 >= x0 && box3->x1 <= x0 + dx
            && box3->y1 - box3->y0 > 19               /* min. length */
            && box3->y1 - box3->y0 >  8 * (bdx)
            && box3->x1 - box3->x0 <  5 * (bdx)
            && ( get_bw(box3->x0+bdx/8 ,box3->x1-bdx/8,
                        box3->y0+bdy/16,box3->y1-bdy/16,box3->p,cs+20,2)==0
                 ||     bdy > 8*bdx
               )
            && ((  abs(box3->y0-by0)<bdy/16+4   /* within bar-box ? */
                && abs(box3->y1-by1)<bdy/4   )  /* lower ends of UPC could be longer */
               ||( abs(box3->y0-yl0)<bdy/16+4   /* left side */
                && abs(box3->y1-yl1)<bdy/4
                && box3->x0 <= bx0           )
               ||( abs(box3->y0-yr0)<bdy/16+4   /* right side */
                && abs(box3->y1-yr1)<bdy/4
                && box3->x0 >= bx1           ) )
            && box3->x0 > bx0 - 12*bbx /* startspace=5...10 */
            && box3->x1 < bx1 + 12*bbx
           ) { /* a bar? */
              if (box3->x0<bx0) { bx0=box3->x0; yl0=box3->y0; yl1=box3->y1; }
              if (box3->x1>bx1) { bx1=box3->x1; yr0=box3->y0; yr1=box3->y1; }
              if (box3->y0<by0) by0=box3->y0; /* overall frame */
              if (box3->y1>by1) by1=box3->y1;
              box3->c=PICTURE; /* BARCODE */
              j++; j2=1;
              list_del(&(JOB->res.boxlist),box3);
              free(box3);
           }
         } end_for_each(&(JOB->res.boxlist));
       }

       if (j>5) {
         box2->c=PICTURE; /* BARCODE */
         box2->x0=bx0;       box2->y0=by0;
         box2->x1=bx1;       box2->y1=by1;
         /* ToDo: add pointer to decoded text */

         y=(box2->y0+box2->y1)/2;
         if (JOB->cfg.verbose){
           if (j!=num_cross(box2->x0,box2->x1,y,y,box2->p,JOB->cfg.cs))
             fprintf(stderr,"# barcode trouble: number of bars was %3d\n",j);
           fprintf(stderr,"# barcode at %d %d size %d %d",
             bx0,by0,bx1-bx0+1,by1-by0+1);
         }
         /* transport the info to the gocr-output (development) */
         box2->obj=(char *)malloc(128);
         if(box2->obj) strncpy(box2->obj,"<!-- barcode? -->",128);
         /* ToDo: decode and print/store barcode bars=j */
         decode_barcode(box2); /* ToDo: char *dest, int len */
         if (JOB->cfg.verbose) fprintf(stderr,"\n");
       }
    }
  } end_for_each(&(JOB->res.boxlist));

  return 0;
}

